from tiders_core import base58_encode_bytes, base58_decode_string

b = b"asdlmsa;dm1123213213:SDMA"

s = base58_encode_bytes(b)

print(s)

b2 = base58_decode_string(s)

print(b)
print(b2)

print(b == b2)



import tiders_core
import polars as pl


def main():
    # Test addresses including invalid and valid ones
    addresses = [
        "Invalid address",
        "0x0000000000000000000000000000000000000000",  # Zero address
        "0x6B175474E89094C44Da98b954EedeAC495271d0F",  # DAI
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",  # USDC
        "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84",  # stETH
    ]

    # Test get_token_metadata
    print("Testing get_token_metadata:")
    token_metadata = tiders_core.get_token_metadata(
        "https://ethereum-rpc.publicnode.com", addresses
    )
    print("Token metadata as list of dictionaries:")
    for metadata in token_metadata:
        print(metadata)
    print("\n")

    # Test get_token_metadata_as_table
    print("Testing get_token_metadata_as_table:")
    token_metadata_table = tiders_core.get_token_metadata_as_table(
        "https://ethereum-rpc.publicnode.com",
        addresses,
        {
            "decimals": True,
            "symbol": False,
            "name": True,
            "total_supply": True,
        },
    )
    # Convert to pandas DataFrame for better display
    df = pl.from_arrow(token_metadata_table)
    print("Token metadata as table:")
    print(df)


if __name__ == "__main__":
    main()



import pyarrow as pa
import pyarrow.parquet as pq
from pathlib import Path
from tiders_core.svm_decode import (
    LogSignature,
    ParamInput,
    DynType,
)
from tiders_core import svm_decode_logs

current_dir = Path(__file__).parent
input_file = current_dir / "logs.parquet"
output_file = current_dir / "decoded_logs.parquet"

print(f"Reading input file: {input_file}")
print(f"Will save output to: {output_file}")

try:
    table = pq.read_table(str(input_file))

    batch = table.to_batches()[0]

    signature = LogSignature(
        params=[
            ParamInput(
                name="whirlpool",
                param_type=DynType.FixedArray(DynType.U8, 32),
            ),
            ParamInput(
                name="a_to_b",
                param_type=DynType.Bool,
            ),
            ParamInput(
                name="pre_sqrt_price",
                param_type=DynType.U128,
            ),
            ParamInput(
                name="post_sqrt_price",
                param_type=DynType.U128,
            ),
            ParamInput(
                name="x",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="input_amount",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="output_amount",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="input_transfer_fee",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="output_transfer_fee",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="lp_fee",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="protocol_fee",
                param_type=DynType.U64,
            ),
        ],
    )

    print("Decoding instruction batch...")
    decoded_batch = svm_decode_logs(signature, batch, True)

    decoded_table = pa.Table.from_batches([decoded_batch])

    print("Saving decoded result...")
    pq.write_table(decoded_table, str(output_file))

    print("Successfully decoded and saved the result!")

except Exception as e:
    print(f"Error during decoding: {e}")
    raise



from cryo import collect as cryo_collect
import tiders_core
import typing
import polars
import pyarrow

signature = "Transfer(address indexed from, address indexed to, uint256 amount)"
topic0 = tiders_core.evm_signature_to_topic0(signature)
contract_address = "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"

# get filtered events from last 10 blocks
data = cryo_collect(
    datatype="logs",
    blocks=["-10:"],
    rpc="https://eth.rpc.hypersync.xyz",
    output_format="polars",
    contract=[contract_address],  # type: ignore[arg-type]
    topic0=[topic0],  # type: ignore[arg-type]
    hex=False,
)

data = typing.cast(polars.DataFrame, data)

batches = data.to_arrow().to_batches()
batch = pyarrow.concat_batches(batches)

# decode events based on the event signature.
# This function automatically infers output types from the signature and can handle arbitrary levels
# of nesting via tuples/lists for example this: https://github.com/yulesa/tiders-core/blob/21534e31ae2e33ae62514765f25d28259ed03129/core/src/tests.rs#L18
decoded = tiders_core.evm_decode_events(signature, batch, allow_decode_fail=False)

# cast to float since polars can't ffi Int256 physical type yet
# https://github.com/pola-rs/polars/blob/main/crates/polars-arrow/src/ffi/array.rs#L26
# https://github.com/pola-rs/polars/blob/main/crates/polars-arrow/src/util/macros.rs#L25
#
# This function is a helper function to do multiple cast operations at once. It casts
# the named column 'amount' to 128 bit integers in this case.
decoded = tiders_core.cast_by_type(
    decoded, pyarrow.decimal256(76, 0), pyarrow.decimal128(38, 0), allow_cast_fail=False
)

# convert all binary columns to prefix hex string format like '0xabc'
decoded = tiders_core.prefix_hex_encode(decoded)

decoded = polars.from_arrow(decoded)
decoded = typing.cast(polars.DataFrame, decoded)
encoded_batch = polars.from_arrow(tiders_core.prefix_hex_encode(batch))
if isinstance(encoded_batch, polars.DataFrame):
    decoded = decoded.hstack(encoded_batch)
else:
    raise ValueError("encoded_batch is not a polars.DataFrame")

print(decoded)

sum = decoded.get_column("amount").sum()
print(f"total volume is {sum}")



from tiders_core import ingest
import asyncio
import argparse


async def run(provider: ingest.ProviderConfig, query: ingest.Query):
    stream = ingest.start_stream(provider, query)

    while True:
        res = await stream.next()
        if res is None:
            break

        print(res["transactions"].schema)


async def main(provider_kind: ingest.ProviderKind):
    query = ingest.Query(
        kind=ingest.QueryKind.EVM,
        params=ingest.evm.Query(
            from_block=0,
            transactions=[ingest.evm.TransactionRequest()],
            fields=ingest.evm.Fields(
                transaction=ingest.evm.TransactionFields(
                    v=True,
                    r=True,
                    s=True,
                ),
            ),
        ),
    )

    url = None

    if provider_kind == ingest.ProviderKind.SQD:
        url = "https://portal.sqd.dev/datasets/ethereum-mainnet"
    elif provider_kind == ingest.ProviderKind.HYPERSYNC:
        url = "https://eth.hypersync.xyz"

    await run(
        ingest.ProviderConfig(
            kind=provider_kind,
            url=url,
            stop_on_head=False,  # default is False as well
            head_poll_interval_millis=1000,  # default is 1000
        ),
        query=query,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="example")

    parser.add_argument(
        "--provider",
        choices=["sqd", "hypersync"],
        required=True,
        help="Specify the provider ('sqd' or 'hypersync')",
    )

    args = parser.parse_args()

    asyncio.run(main(args.provider))



from tiders_core import ingest
import pyarrow.parquet as pq
import pyarrow as pa
import asyncio


async def run(provider: ingest.ProviderConfig, query: ingest.Query):
    stream = ingest.start_stream(provider, query)

    while True:
        res = await stream.next()
        if res is None:
            break

        print(res.keys())

        for key in res.keys():
            table = res[key]
            table = pa.Table.from_batches([table])
            pq.write_table(table, f"data/{key}_{start_block}_{end_block}.parquet")


start_block = 317617480
end_block = 317617580
url = "https://portal.sqd.dev/datasets/solana-mainnet"

query = ingest.Query(
    kind=ingest.QueryKind.SVM,
    params=ingest.svm.Query(
        from_block=start_block,
        to_block=end_block,
        include_all_blocks=True,
        fields=ingest.svm.Fields(
            block=ingest.svm.BlockFields(
                slot=True,
                hash=True,
                parent_slot=True,
                parent_hash=True,
                height=True,
                timestamp=True,
            ),
            instruction=ingest.svm.InstructionFields(
                block_slot=True,
                block_hash=True,
                transaction_index=True,
                instruction_address=True,
                program_id=True,
                a0=True,
                a1=True,
                a2=True,
                a3=True,
                a4=True,
                a5=True,
                a6=True,
                a7=True,
                a8=True,
                a9=True,
                rest_of_accounts=True,
                data=True,
                d1=True,
                d2=True,
                d4=True,
                d8=True,
                error=True,
                compute_units_consumed=True,
                is_committed=True,
                has_dropped_log_messages=True,
            ),
            transaction=ingest.svm.TransactionFields(
                block_slot=True,
                block_hash=True,
                transaction_index=True,
                signature=True,
                version=True,
                account_keys=True,
                address_table_lookups=True,
                num_readonly_signed_accounts=True,
                num_readonly_unsigned_accounts=True,
                num_required_signatures=True,
                # recent_blockhash=True,
                signatures=True,
                err=True,
                fee=True,
                compute_units_consumed=True,
                loaded_readonly_addresses=True,
                loaded_writable_addresses=True,
                fee_payer=True,
                has_dropped_log_messages=True,
            ),
            log=ingest.svm.LogFields(
                block_slot=True,
                block_hash=True,
                transaction_index=True,
                log_index=True,
                instruction_address=True,
                program_id=True,
                kind=True,
                message=True,
            ),
            balance=ingest.svm.BalanceFields(
                block_slot=True,
                block_hash=True,
                transaction_index=True,
                account=True,
                pre=True,
                post=True,
            ),
            token_balance=ingest.svm.TokenBalanceFields(
                block_slot=True,
                block_hash=True,
                transaction_index=True,
                account=True,
                pre_mint=True,
                post_mint=True,
                pre_decimals=True,
                post_decimals=True,
                pre_program_id=True,
                post_program_id=True,
                pre_owner=True,
                post_owner=True,
                pre_amount=True,
                post_amount=True,
            ),
            reward=ingest.svm.RewardFields(
                block_slot=True,
                block_hash=True,
                pubkey=True,
                lamports=True,
                post_balance=True,
                reward_type=True,
                commission=True,
            ),
        ),
        instructions=[ingest.svm.InstructionRequest()],
        transactions=[ingest.svm.TransactionRequest()],
        logs=[ingest.svm.LogRequest()],
        balances=[ingest.svm.BalanceRequest()],
        token_balances=[ingest.svm.TokenBalanceRequest()],
        rewards=[ingest.svm.RewardRequest()],
    ),
)

print("running with sqd")
asyncio.run(
    run(
        ingest.ProviderConfig(
            kind=ingest.ProviderKind.SQD,
            url=url,
        ),
        query=query,
    )
)



import tiders_core
from tiders_core import ingest
import asyncio
import logging
import os

logging.basicConfig(level=os.environ.get("LOGLEVEL", "DEBUG").upper())
logger = logging.getLogger(__name__)

signature = "Transfer(address indexed from, address indexed to, uint256 amount)"
topic0 = tiders_core.evm_signature_to_topic0(signature)
contract_address = "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"


async def run(provider: ingest.ProviderConfig, query: ingest.Query):
    stream = ingest.start_stream(provider, query)

    while True:
        res = await stream.next()
        if res is None:
            break

        logger.info(res["blocks"].column("number"))
        logger.debug(res)

        decoded = tiders_core.evm_decode_events(signature, res["logs"])
        logger.debug(decoded)


query = ingest.Query(
    kind=ingest.QueryKind.EVM,
    params=ingest.evm.Query(
        from_block=20123123,
        to_block=20123223,
        logs=[
            ingest.evm.LogRequest(
                address=[contract_address],
                topic0=[topic0],
                include_blocks=True,
            )
        ],
        fields=ingest.evm.Fields(
            block=ingest.evm.BlockFields(
                number=True,
            ),
            log=ingest.evm.LogFields(
                data=True,
                topic0=True,
                topic1=True,
                topic2=True,
                topic3=True,
            ),
        ),
    ),
)

print("running with sqd")
asyncio.run(
    run(
        ingest.ProviderConfig(
            kind=ingest.ProviderKind.SQD,
            url="https://portal.sqd.dev/datasets/ethereum-mainnet",
        ),
        query=query,
    )
)

print("running with hypersync")
asyncio.run(
    run(
        ingest.ProviderConfig(
            kind=ingest.ProviderKind.HYPERSYNC,
            url="https://eth.hypersync.xyz",
        ),
        query=query,
    )
)



from tiders_core import ingest, base58_encode
import asyncio
import polars
from typing import cast
import pyarrow as pa
import os


async def run(provider: ingest.ProviderConfig, query: ingest.Query):
    stream = ingest.start_stream(provider, query)

    while True:
        res = await stream.next()
        if res is None:
            break

        transactions = cast(polars.DataFrame, polars.from_arrow(res["transactions"]))
        token_balances = cast(
            polars.DataFrame, polars.from_arrow(res["token_balances"])
        )

        token_balances = token_balances.join(
            transactions, ["block_slot", "transaction_index"]
        )

        for batch in token_balances.to_arrow().to_batches():
            new_batch = batch
            for i, col in enumerate(new_batch.columns):
                if pa.types.is_large_binary(col.type):
                    new_batch = new_batch.set_column(
                        i, new_batch.column_names[i], col.cast(target_type=pa.binary())
                    )

            new_batch = base58_encode(new_batch)

            print(new_batch)


query = ingest.Query(
    kind=ingest.QueryKind.SVM,
    params=ingest.svm.Query(
        from_block=317617480,
        token_balances=[
            ingest.svm.TokenBalanceRequest(
                include_transactions=True,
            )
        ],
        fields=ingest.svm.Fields(
            token_balance=ingest.svm.TokenBalanceFields(
                block_slot=True,
                transaction_index=True,
                post_owner=True,
                pre_owner=True,
                post_amount=True,
                pre_amount=True,
                post_mint=True,
                account=True,
            ),
            transaction=ingest.svm.TransactionFields(
                block_slot=True,
                transaction_index=True,
                signature=True,
                err=True,
            ),
        ),
    ),
)

asyncio.run(
    run(
        ingest.ProviderConfig(
            kind=ingest.ProviderKind.YELLOWSTONE_GRPC,
            url=os.environ.get("YELLOWSTONE_GRPC_URL"),
            bearer_token=os.environ.get("YELLOWSTONE_GRPC_TOKEN"),
        ),
        query=query,
    )
)



import pyarrow as pa
import pyarrow.parquet as pq
from pathlib import Path
from tiders_core.svm_decode import (
    InstructionSignature,
    ParamInput,
    DynType,
    Field,
    Variant,
)
from tiders_core import svm_decode_instructions

current_dir = Path(__file__).parent
input_file = current_dir / "instructions.parquet"
output_file = current_dir / "decoded.parquet"

print(f"Reading input file: {input_file}")
print(f"Will save output to: {output_file}")

try:
    table = pq.read_table(str(input_file))

    batch = table.to_batches()[0]

    signature = InstructionSignature(
        discriminator="e517cb977ae3ad2a",
        params=[
            ParamInput(
                name="RoutePlan",
                param_type=DynType.Array(
                    DynType.Struct(
                        [
                            Field(
                                name="Swap",
                                element_type=DynType.Enum(
                                    [
                                        Variant("Saber", None),
                                        Variant("SaberAddDecimalsDeposit", None),
                                        Variant("SaberAddDecimalsWithdraw", None),
                                        Variant("TokenSwap", None),
                                        Variant("Sencha", None),
                                        Variant("Step", None),
                                        Variant("Cropper", None),
                                        Variant("Raydium", None),
                                        Variant(
                                            "Crema",
                                            DynType.Struct(
                                                [
                                                    Field("a_to_b", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant("Lifinity", None),
                                        Variant("Mercurial", None),
                                        Variant("Cykura", None),
                                        Variant(
                                            "Serum",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("MarinadeDeposit", None),
                                        Variant("MarinadeUnstake", None),
                                        Variant(
                                            "Aldrin",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "AldrinV2",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "Whirlpool",
                                            DynType.Struct(
                                                [
                                                    Field("a_to_b", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "Invariant",
                                            DynType.Struct(
                                                [
                                                    Field("x_to_y", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant("Meteora", None),
                                        Variant("GooseFX", None),
                                        Variant(
                                            "DeltaFi",
                                            DynType.Struct(
                                                [
                                                    Field("stable", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant("Balansol", None),
                                        Variant(
                                            "MarcoPolo",
                                            DynType.Struct(
                                                [
                                                    Field("x_to_y", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "Dradex",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("LifinityV2", None),
                                        Variant("RaydiumClmm", None),
                                        Variant(
                                            "Openbook",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "Phoenix",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "Symmetry",
                                            DynType.Struct(
                                                [
                                                    Field("from_token_id", DynType.U64),
                                                    Field("to_token_id", DynType.U64),
                                                ]
                                            ),
                                        ),
                                        Variant("TokenSwapV2", None),
                                        Variant(
                                            "HeliumTreasuryManagementRedeemV0", None
                                        ),
                                        Variant("StakeDexStakeWrappedSol", None),
                                        Variant(
                                            "StakeDexSwapViaStake",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "bridge_stake_seed", DynType.U32
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("GooseFXV2", None),
                                        Variant("Perps", None),
                                        Variant("PerpsAddLiquidity", None),
                                        Variant("PerpsRemoveLiquidity", None),
                                        Variant("MeteoraDlmm", None),
                                        Variant(
                                            "OpenBookV2",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("RaydiumClmmV2", None),
                                        Variant(
                                            "StakeDexPrefundWithdrawStakeAndDepositStake",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "bridge_stake_seed", DynType.U32
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "Clone",
                                            DynType.Struct(
                                                [
                                                    Field("pool_index", DynType.U8),
                                                    Field(
                                                        "quantity_is_input",
                                                        DynType.Bool,
                                                    ),
                                                    Field(
                                                        "quantity_is_collateral",
                                                        DynType.Bool,
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "SanctumS",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "src_lst_value_calc_accs",
                                                        DynType.U8,
                                                    ),
                                                    Field(
                                                        "dst_lst_value_calc_accs",
                                                        DynType.U8,
                                                    ),
                                                    Field("src_lst_index", DynType.U32),
                                                    Field("dst_lst_index", DynType.U32),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "SanctumSAddLiquidity",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "lst_value_calc_accs",
                                                        DynType.U8,
                                                    ),
                                                    Field("lst_index", DynType.U32),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "SanctumSRemoveLiquidity",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "lst_value_calc_accs",
                                                        DynType.U8,
                                                    ),
                                                    Field("lst_index", DynType.U32),
                                                ]
                                            ),
                                        ),
                                        Variant("RaydiumCP", None),
                                        Variant(
                                            "WhirlpoolSwapV2",
                                            DynType.Struct(
                                                [
                                                    Field("a_to_b", DynType.Bool),
                                                    Field(
                                                        "remaining_accounts_info",
                                                        DynType.Struct(
                                                            [
                                                                Field(
                                                                    "slices",
                                                                    DynType.Array(
                                                                        DynType.Struct(
                                                                            [
                                                                                Field(
                                                                                    "remaining_accounts_slice",
                                                                                    DynType.Struct(
                                                                                        [
                                                                                            Field(
                                                                                                "accounts_type",
                                                                                                DynType.U8,
                                                                                            ),
                                                                                            Field(
                                                                                                "length",
                                                                                                DynType.U8,
                                                                                            ),
                                                                                        ]
                                                                                    ),
                                                                                )
                                                                            ]
                                                                        )
                                                                    ),
                                                                ),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("OneIntro", None),
                                        Variant("PumpdotfunWrappedBuy", None),
                                        Variant("PumpdotfunWrappedSell", None),
                                        Variant("PerpsV2", None),
                                        Variant("PerpsV2AddLiquidity", None),
                                        Variant("PerpsV2RemoveLiquidity", None),
                                        Variant("MoonshotWrappedBuy", None),
                                        Variant("MoonshotWrappedSell", None),
                                        Variant("StabbleStableSwap", None),
                                        Variant("StabbleWeightedSwap", None),
                                        Variant(
                                            "Obric",
                                            DynType.Struct(
                                                [
                                                    Field("x_to_y", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant("FoxBuyFromEstimatedCost", None),
                                        Variant(
                                            "FoxClaimPartial",
                                            DynType.Struct(
                                                [
                                                    Field("is_y", DynType.Bool),
                                                ]
                                            ),
                                        ),
                                        Variant(
                                            "SolFi",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "is_quote_to_base", DynType.Bool
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("SolayerDelegateNoInit", None),
                                        Variant("SolayerUndelegateNoInit", None),
                                        Variant(
                                            "TokenMill",
                                            DynType.Struct(
                                                [
                                                    Field(
                                                        "side",
                                                        DynType.Enum(
                                                            [
                                                                Variant("Bid", None),
                                                                Variant("Ask", None),
                                                            ]
                                                        ),
                                                    ),
                                                ]
                                            ),
                                        ),
                                        Variant("DaosFunBuy", None),
                                        Variant("DaosFunSell", None),
                                        Variant("ZeroFi", None),
                                        Variant("StakeDexWithdrawWrappedSol", None),
                                        Variant("VirtualsBuy", None),
                                        Variant("VirtualsSell", None),
                                        Variant(
                                            "Peren",
                                            DynType.Struct(
                                                [
                                                    Field("in_index", DynType.U8),
                                                    Field("out_index", DynType.U8),
                                                ]
                                            ),
                                        ),
                                        Variant("PumpdotfunAmmBuy", None),
                                        Variant("PumpdotfunAmmSell", None),
                                        Variant("Gamma", None),
                                    ]
                                ),
                            ),
                            Field("Percent", DynType.U8),
                            Field("InputIndex", DynType.U8),
                            Field("OutputIndex", DynType.U8),
                        ]
                    )
                ),
            ),
            ParamInput(
                name="InAmount",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="QuotedOutAmount",
                param_type=DynType.U64,
            ),
            ParamInput(
                name="SlippageBps",
                param_type=DynType.U16,
            ),
            ParamInput(
                name="PlatformFeeBps",
                param_type=DynType.U8,
            ),
        ],
        accounts_names=[
            "TokenProgram",
            "UserTransferAuthority",
            "UserSourceTokenAccount",
            "UserDestinationTokenAccount",
            "DestinationTokenAccount",
            "PlatformFeeAccount",
            "EventAuthority",
            "Program",
        ],
    )

    print("Decoding instruction batch...")
    decoded_batch = svm_decode_instructions(signature, batch, True)

    decoded_table = pa.Table.from_batches([decoded_batch])

    print("Saving decoded result...")
    pq.write_table(decoded_table, str(output_file))

    print("Successfully decoded and saved the result!")

except Exception as e:
    print(f"Error during decoding: {e}")
    raise



import tiders_core
import polars as pl


def main():
    # Test addresses including invalid and valid ones
    pool_addresses = [
        "Invalid address",
        "0xfBB6Eed8e7aa03B138556eeDaF5D271A5E1e43ef",  # cbBTC/USDC on uniswap v3
        "0x31f609019d0CC0b8cC865656142d6FeD69853689",  # POPCAT/WETH on uniswap v2
        "0x6cDcb1C4A4D1C3C6d054b27AC5B77e89eAFb971d",  # AERO/USDC on Aerodrome
        "0x323b43332F97B1852D8567a08B1E8ed67d25A8d5",  # msETH/WETH on Pancake Swap
    ]

    # Test get_pools_token0_token1
    print("Testing get_pools_token0_token1:")
    pool_tokens = tiders_core.get_pools_token0_token1(
        "https://base-rpc.publicnode.com", pool_addresses
    )
    print("Pool tokens as list of dictionaries:")
    for pool in pool_tokens:
        print(pool)
    print("\n")

    # Test get_pools_token0_token1_as_table
    print("Testing get_pools_token0_token1_as_table:")
    pool_tokens_table = tiders_core.get_pools_token0_token1_as_table(
        "https://base-rpc.publicnode.com", pool_addresses
    )
    # Convert to polars DataFrame for better display
    df = pl.from_arrow(pool_tokens_table)
    print("Pool tokens as table:")
    print(df)


if __name__ == "__main__":
    main()



from tiders_core import ingest
import asyncio


async def run(provider: ingest.ProviderConfig, query: ingest.Query):
    stream = ingest.start_stream(provider, query)

    while True:
        res = await stream.next()
        if res is None:
            break

        print(res["blocks"].column("number"))


query = ingest.Query(
    kind=ingest.QueryKind.EVM,
    params=ingest.evm.Query(
        from_block=21930160,
        include_all_blocks=True,
        fields=ingest.evm.Fields(
            block=ingest.evm.BlockFields(
                number=True,
            ),
        ),
    ),
)

asyncio.run(
    run(
        ingest.ProviderConfig(
            kind=ingest.ProviderKind.SQD,
            url="https://portal.sqd.dev/datasets/ethereum-mainnet",
            stop_on_head=False,  # default is False as well
            head_poll_interval_millis=1000,  # default is 1000
        ),
        query=query,
    )
)



from typing import Optional
from dataclasses import dataclass, field


@dataclass
class TransactionRequest:
    from_: list[str] = field(default_factory=list)
    to: list[str] = field(default_factory=list)
    sighash: list[str] = field(default_factory=list)
    status: list[int] = field(default_factory=list)
    type_: list[int] = field(default_factory=list)
    contract_deployment_address: list[str] = field(default_factory=list)
    hash: list[str] = field(default_factory=list)
    include_logs: bool = False
    include_traces: bool = False
    include_blocks: bool = False


@dataclass
class LogRequest:
    address: list[str] = field(default_factory=list)
    topic0: list[str] = field(default_factory=list)
    topic1: list[str] = field(default_factory=list)
    topic2: list[str] = field(default_factory=list)
    topic3: list[str] = field(default_factory=list)
    include_transactions: bool = False
    include_transaction_logs: bool = False
    include_transaction_traces: bool = False
    include_blocks: bool = False


@dataclass
class TraceRequest:
    from_: list[str] = field(default_factory=list)
    to: list[str] = field(default_factory=list)
    address: list[str] = field(default_factory=list)
    call_type: list[str] = field(default_factory=list)
    reward_type: list[str] = field(default_factory=list)
    type_: list[str] = field(default_factory=list)
    sighash: list[str] = field(default_factory=list)
    author: list[str] = field(default_factory=list)
    include_transactions: bool = False
    include_transaction_logs: bool = False
    include_transaction_traces: bool = False
    include_blocks: bool = False


@dataclass
class BlockFields:
    number: bool = False
    hash: bool = False
    parent_hash: bool = False
    nonce: bool = False
    sha3_uncles: bool = False
    logs_bloom: bool = False
    transactions_root: bool = False
    state_root: bool = False
    receipts_root: bool = False
    miner: bool = False
    difficulty: bool = False
    total_difficulty: bool = False
    extra_data: bool = False
    size: bool = False
    gas_limit: bool = False
    gas_used: bool = False
    timestamp: bool = False
    uncles: bool = False
    base_fee_per_gas: bool = False
    blob_gas_used: bool = False
    excess_blob_gas: bool = False
    parent_beacon_block_root: bool = False
    withdrawals_root: bool = False
    withdrawals: bool = False
    l1_block_number: bool = False
    send_count: bool = False
    send_root: bool = False
    mix_hash: bool = False


@dataclass
class TransactionFields:
    block_hash: bool = False
    block_number: bool = False
    from_: bool = False
    gas: bool = False
    gas_price: bool = False
    hash: bool = False
    input: bool = False
    nonce: bool = False
    to: bool = False
    transaction_index: bool = False
    value: bool = False
    v: bool = False
    r: bool = False
    s: bool = False
    max_priority_fee_per_gas: bool = False
    max_fee_per_gas: bool = False
    chain_id: bool = False
    cumulative_gas_used: bool = False
    effective_gas_price: bool = False
    gas_used: bool = False
    contract_address: bool = False
    logs_bloom: bool = False
    type_: bool = False
    root: bool = False
    status: bool = False
    sighash: bool = False
    y_parity: bool = False
    access_list: bool = False
    l1_fee: bool = False
    l1_gas_price: bool = False
    l1_fee_scalar: bool = False
    gas_used_for_l1: bool = False
    max_fee_per_blob_gas: bool = False
    blob_versioned_hashes: bool = False
    deposit_nonce: bool = False
    blob_gas_price: bool = False
    deposit_receipt_version: bool = False
    blob_gas_used: bool = False
    l1_base_fee_scalar: bool = False
    l1_blob_base_fee: bool = False
    l1_blob_base_fee_scalar: bool = False
    l1_block_number: bool = False
    mint: bool = False
    source_hash: bool = False


@dataclass
class LogFields:
    removed: bool = False
    log_index: bool = False
    transaction_index: bool = False
    transaction_hash: bool = False
    block_hash: bool = False
    block_number: bool = False
    address: bool = False
    data: bool = False
    topic0: bool = False
    topic1: bool = False
    topic2: bool = False
    topic3: bool = False


@dataclass
class TraceFields:
    from_: bool = False
    to: bool = False
    call_type: bool = False
    gas: bool = False
    input: bool = False
    init: bool = False
    value: bool = False
    author: bool = False
    reward_type: bool = False
    block_hash: bool = False
    block_number: bool = False
    address: bool = False
    code: bool = False
    gas_used: bool = False
    output: bool = False
    subtraces: bool = False
    trace_address: bool = False
    transaction_hash: bool = False
    transaction_position: bool = False
    type_: bool = False
    error: bool = False
    sighash: bool = False
    action_address: bool = False
    balance: bool = False
    refund_address: bool = False


@dataclass
class Fields:
    block: BlockFields = field(default_factory=BlockFields)
    transaction: TransactionFields = field(default_factory=TransactionFields)
    log: LogFields = field(default_factory=LogFields)
    trace: TraceFields = field(default_factory=TraceFields)


@dataclass
class Query:
    from_block: int = 0
    to_block: Optional[int] = None
    include_all_blocks: bool = False
    transactions: list[TransactionRequest] = field(default_factory=list)
    logs: list[LogRequest] = field(default_factory=list)
    traces: list[TraceRequest] = field(default_factory=list)
    fields: Fields = field(default_factory=Fields)



from typing import List, Optional, Union, TypeAlias, Literal
from dataclasses import dataclass

PrimitiveType: TypeAlias = Literal[
    "i8", "i16", "i32", "i64", "i128", "u8", "u16", "u32", "u64", "u128", "bool"
]
ElementType: TypeAlias = Union[
    PrimitiveType, "DynType", "FixedArray", "Array", "Struct", "Enum", "Option"
]


@dataclass
class FixedArray:
    element_type: ElementType
    size: int


@dataclass
class Array:
    element_type: ElementType


@dataclass
class Field:
    name: str
    element_type: ElementType


@dataclass
class Struct:
    fields: List[Field]


@dataclass
class Variant:
    name: str
    element_type: Optional[ElementType]


@dataclass
class Enum:
    variants: List[Variant]


@dataclass
class Option:
    element_type: ElementType


@dataclass
class ParamInput:
    name: str
    param_type: ElementType


@dataclass
class InstructionSignature:
    discriminator: Union[bytes, str]
    params: List[ParamInput]
    accounts_names: List[str]


@dataclass
class LogSignature:
    params: List[ParamInput]


class DynType:
    I8: PrimitiveType = "i8"
    I16: PrimitiveType = "i16"
    I32: PrimitiveType = "i32"
    I64: PrimitiveType = "i64"
    I128: PrimitiveType = "i128"
    U8: PrimitiveType = "u8"
    U16: PrimitiveType = "u16"
    U32: PrimitiveType = "u32"
    U64: PrimitiveType = "u64"
    U128: PrimitiveType = "u128"
    Bool: PrimitiveType = "bool"
    FixedArray = FixedArray
    Array = Array
    Struct = Struct
    Enum = Enum
    Option = Option



from typing import Optional, Union
from dataclasses import dataclass, field
from enum import Enum


@dataclass
class InstructionFields:
    block_slot: bool = False
    block_hash: bool = False
    transaction_index: bool = False
    instruction_address: bool = False
    program_id: bool = False
    a0: bool = False
    a1: bool = False
    a2: bool = False
    a3: bool = False
    a4: bool = False
    a5: bool = False
    a6: bool = False
    a7: bool = False
    a8: bool = False
    a9: bool = False
    rest_of_accounts: bool = False
    data: bool = False
    d1: bool = False
    d2: bool = False
    d4: bool = False
    d8: bool = False
    error: bool = False
    compute_units_consumed: bool = False
    is_committed: bool = False
    has_dropped_log_messages: bool = False


@dataclass
class TransactionFields:
    block_slot: bool = False
    block_hash: bool = False
    transaction_index: bool = False
    signature: bool = False
    version: bool = False
    account_keys: bool = False
    address_table_lookups: bool = False
    num_readonly_signed_accounts: bool = False
    num_readonly_unsigned_accounts: bool = False
    num_required_signatures: bool = False
    recent_blockhash: bool = False
    signatures: bool = False
    err: bool = False
    fee: bool = False
    compute_units_consumed: bool = False
    loaded_readonly_addresses: bool = False
    loaded_writable_addresses: bool = False
    fee_payer: bool = False
    has_dropped_log_messages: bool = False


@dataclass
class LogFields:
    block_slot: bool = False
    block_hash: bool = False
    transaction_index: bool = False
    log_index: bool = False
    instruction_address: bool = False
    program_id: bool = False
    kind: bool = False
    message: bool = False


@dataclass
class BalanceFields:
    block_slot: bool = False
    block_hash: bool = False
    transaction_index: bool = False
    account: bool = False
    pre: bool = False
    post: bool = False


@dataclass
class TokenBalanceFields:
    block_slot: bool = False
    block_hash: bool = False
    transaction_index: bool = False
    account: bool = False
    pre_mint: bool = False
    post_mint: bool = False
    pre_decimals: bool = False
    post_decimals: bool = False
    pre_program_id: bool = False
    post_program_id: bool = False
    pre_owner: bool = False
    post_owner: bool = False
    pre_amount: bool = False
    post_amount: bool = False


@dataclass
class RewardFields:
    block_slot: bool = False
    block_hash: bool = False
    pubkey: bool = False
    lamports: bool = False
    post_balance: bool = False
    reward_type: bool = False
    commission: bool = False


@dataclass
class BlockFields:
    slot: bool = False
    hash: bool = False
    parent_slot: bool = False
    parent_hash: bool = False
    height: bool = False
    timestamp: bool = False


@dataclass
class Fields:
    instruction: InstructionFields = field(default_factory=InstructionFields)
    transaction: TransactionFields = field(default_factory=TransactionFields)
    log: LogFields = field(default_factory=LogFields)
    balance: BalanceFields = field(default_factory=BalanceFields)
    token_balance: TokenBalanceFields = field(default_factory=TokenBalanceFields)
    reward: RewardFields = field(default_factory=RewardFields)
    block: BlockFields = field(default_factory=BlockFields)


@dataclass
class InstructionRequest:
    program_id: list[str] = field(default_factory=list)
    discriminator: list[Union[bytes, str]] = field(default_factory=list)
    d1: list[Union[bytes, str]] = field(default_factory=list)
    d2: list[Union[bytes, str]] = field(default_factory=list)
    d3: list[Union[bytes, str]] = field(default_factory=list)
    d4: list[Union[bytes, str]] = field(default_factory=list)
    d8: list[Union[bytes, str]] = field(default_factory=list)
    a0: list[str] = field(default_factory=list)
    a1: list[str] = field(default_factory=list)
    a2: list[str] = field(default_factory=list)
    a3: list[str] = field(default_factory=list)
    a4: list[str] = field(default_factory=list)
    a5: list[str] = field(default_factory=list)
    a6: list[str] = field(default_factory=list)
    a7: list[str] = field(default_factory=list)
    a8: list[str] = field(default_factory=list)
    a9: list[str] = field(default_factory=list)
    is_committed: bool = False
    include_transactions: bool = False
    include_transaction_token_balances: bool = False
    include_logs: bool = False
    include_inner_instructions: bool = False
    include_blocks: bool = True


@dataclass
class TransactionRequest:
    fee_payer: list[str] = field(default_factory=list)
    include_instructions: bool = False
    include_logs: bool = False
    include_blocks: bool = False


class LogKind(str, Enum):
    LOG = "log"
    DATA = "data"
    OTHER = "other"


@dataclass
class LogRequest:
    program_id: list[str] = field(default_factory=list)
    kind: list[LogKind] = field(
        default_factory=list
    )  # Assuming LogKind is represented as a string
    include_transactions: bool = False
    include_instructions: bool = False
    include_blocks: bool = False


@dataclass
class BalanceRequest:
    account: list[str] = field(default_factory=list)
    include_transactions: bool = False
    include_transaction_instructions: bool = False
    include_blocks: bool = False


@dataclass
class TokenBalanceRequest:
    account: list[str] = field(default_factory=list)
    pre_program_id: list[str] = field(default_factory=list)
    post_program_id: list[str] = field(default_factory=list)
    pre_mint: list[str] = field(default_factory=list)
    post_mint: list[str] = field(default_factory=list)
    pre_owner: list[str] = field(default_factory=list)
    post_owner: list[str] = field(default_factory=list)
    include_transactions: bool = False
    include_transaction_instructions: bool = False
    include_blocks: bool = False


@dataclass
class RewardRequest:
    pubkey: list[str] = field(default_factory=list)
    include_blocks: bool = False


@dataclass
class Query:
    from_block: int = 0
    to_block: Optional[int] = None
    include_all_blocks: bool = False
    fields: Fields = field(default_factory=Fields)
    instructions: list[InstructionRequest] = field(default_factory=list)
    transactions: list[TransactionRequest] = field(default_factory=list)
    logs: list[LogRequest] = field(default_factory=list)
    balances: list[BalanceRequest] = field(default_factory=list)
    token_balances: list[TokenBalanceRequest] = field(default_factory=list)
    rewards: list[RewardRequest] = field(default_factory=list)



from tiders_core import ingest, base58_encode_bytes
import asyncio


async def run(provider: ingest.ProviderConfig, query: ingest.Query):
    stream = ingest.start_stream(provider, query)

    while True:
        res = await stream.next()
        if res is None:
            break

        # print(res)

        print(res["blocks"].column("slot"))

        for x in res["transactions"].column("signature"):
            print(base58_encode_bytes(x.as_py()))

        print(res["instructions"].column("instruction_address"))
        print(res["instructions"].column("transaction_index"))


query = ingest.Query(
    kind=ingest.QueryKind.SVM,
    params=ingest.svm.Query(
        from_block=332557668,
        to_block=332557668,
        include_all_blocks=True,
        fields=ingest.svm.Fields(
            block=ingest.svm.BlockFields(
                slot=True,
                hash=True,
            ),
            instruction=ingest.svm.InstructionFields(
                program_id=True,
                data=True,
                instruction_address=True,
                transaction_index=True,
            ),
            transaction=ingest.svm.TransactionFields(
                signature=True,
                transaction_index=True,
            ),
        ),
        instructions=[
            ingest.svm.InstructionRequest(
                program_id=["whirLbMiicVdio4qvUfM5KAg6Ct8VwpYzGff3uctyCc"],
                discriminator=["0xf8c6"],
                a1=["EV5Xoy9TQc4zXtqRHpjDefrxsQTi4Lm12KAE8axyBsdp"],
                # discriminator=[bytes([2, 0, 0, 0, 1, 0, 0, 0])],
                include_inner_instructions=True,
                include_transactions=True,
                is_committed=True,
            )
        ],
    ),
)

print("running with sqd")
asyncio.run(
    run(
        ingest.ProviderConfig(
            kind=ingest.ProviderKind.SQD,
            url="https://portal.sqd.dev/datasets/solana-beta",
        ),
        query=query,
    )
)



import tiders_core.tiders_core as cc
from . import svm_decode
from typing import Tuple
import pyarrow


def cast(
    map: list[Tuple[str, pyarrow.DataType]],
    data: pyarrow.RecordBatch,
    allow_cast_fail: bool = False,
) -> pyarrow.RecordBatch:
    return cc.cast(map, data, allow_cast_fail)


def cast_schema(
    map: list[Tuple[str, pyarrow.DataType]], schema: pyarrow.Schema
) -> pyarrow.Schema:
    return cc.cast_schema(map, schema)


def cast_by_type(
    data: pyarrow.RecordBatch,
    from_type: pyarrow.DataType,
    to_type: pyarrow.DataType,
    allow_cast_fail: bool = False,
) -> pyarrow.RecordBatch:
    return cc.cast_by_type(data, from_type, to_type, allow_cast_fail)


def cast_schema_by_type(
    schema: pyarrow.Schema, from_type: pyarrow.DataType, to_type: pyarrow.DataType
) -> pyarrow.Schema:
    return cc.cast_schema_by_type(schema, from_type, to_type)


def base58_encode(data: pyarrow.RecordBatch) -> pyarrow.RecordBatch:
    return cc.base58_encode(data)


def hex_encode(data: pyarrow.RecordBatch) -> pyarrow.RecordBatch:
    return cc.hex_encode(data)


def prefix_hex_encode(data: pyarrow.RecordBatch) -> pyarrow.RecordBatch:
    return cc.prefix_hex_encode(data)


def base58_encode_column(col: pyarrow.Array) -> pyarrow.Array:
    return cc.base58_encode_column(col)


def hex_encode_column(col: pyarrow.Array) -> pyarrow.Array:
    return cc.hex_encode_column(col)


def prefix_hex_encode_column(col: pyarrow.Array) -> pyarrow.Array:
    return cc.prefix_hex_encode_column(col)


def base58_decode_column(col: pyarrow.Array) -> pyarrow.Array:
    return cc.base58_decode_column(col)


def hex_decode_column(col: pyarrow.Array) -> pyarrow.Array:
    return cc.hex_decode_column(col)


def prefix_hex_decode_column(col: pyarrow.Array) -> pyarrow.Array:
    return cc.prefix_hex_decode_column(col)


def u256_column_from_binary(col: pyarrow.Array) -> pyarrow.Array:
    return cc.u256_column_from_binary(col)


def u256_column_to_binary(col: pyarrow.Array) -> pyarrow.Array:
    return cc.u256_column_to_binary(col)


def u256_to_binary(data: pyarrow.RecordBatch) -> pyarrow.RecordBatch:
    return cc.u256_to_binary(data)


def svm_decode_instructions(
    signature: svm_decode.InstructionSignature,
    batch: pyarrow.RecordBatch,
    allow_decode_fail: bool = False,
) -> pyarrow.RecordBatch:
    return cc.svm_decode_instructions(signature, batch, allow_decode_fail)


def svm_decode_logs(
    signature: svm_decode.LogSignature,
    batch: pyarrow.RecordBatch,
    allow_decode_fail: bool = False,
) -> pyarrow.RecordBatch:
    return cc.svm_decode_logs(signature, batch, allow_decode_fail)


def instruction_signature_to_arrow_schema(
    signature: svm_decode.InstructionSignature,
) -> pyarrow.Schema:
    return cc.instruction_signature_to_arrow_schema(signature)


def evm_decode_call_inputs(
    signature: str, data: pyarrow.Array, allow_decode_fail: bool = False
) -> pyarrow.RecordBatch:
    return cc.evm_decode_call_inputs(signature, data, allow_decode_fail)


def evm_decode_call_outputs(
    signature: str, data: pyarrow.Array, allow_decode_fail: bool = False
) -> pyarrow.RecordBatch:
    return cc.evm_decode_call_outputs(signature, data, allow_decode_fail)


def evm_decode_events(
    signature: str, data: pyarrow.RecordBatch, allow_decode_fail: bool = False
) -> pyarrow.RecordBatch:
    return cc.evm_decode_events(signature, data, allow_decode_fail)


def evm_event_signature_to_arrow_schema(signature: str) -> pyarrow.Schema:
    return cc.evm_event_signature_to_arrow_schema(signature)


def evm_transaction_signature_to_arrow_schemas(
    signature: str,
) -> Tuple[pyarrow.Schema, pyarrow.Schema]:
    return cc.evm_transaction_signature_to_arrow_schemas(signature)


def evm_signature_to_topic0(signature: str) -> str:
    return cc.evm_signature_to_topic0(signature)


def base58_encode_bytes(b: bytes) -> str:
    return cc.base58_encode_bytes(b)


def base58_decode_string(s: str) -> bytes:
    return cc.base58_decode_string(s)



from typing import Dict, Optional
import dataclasses
from dataclasses import dataclass
import pyarrow
from enum import Enum
from . import evm, svm
import tiders_core.tiders_core as cc


class ProviderKind(str, Enum):
    SQD = "sqd"
    HYPERSYNC = "hypersync"
    YELLOWSTONE_GRPC = "yellowstone_grpc"
    RPC = "rpc"


class QueryKind(str, Enum):
    EVM = "evm"
    SVM = "svm"


@dataclass
class Query:
    kind: QueryKind
    params: evm.Query | svm.Query


@dataclass
class ProviderConfig:
    kind: ProviderKind  # (Required) The selected provider.
    url: Optional[str] = None  # (Optional) The URL of the provider.
    bearer_token: Optional[str] = None  # (Optional) Optional authentication token for protected APIs.
    max_num_retries: Optional[int] = None  # (Optional) Maximum number of retries for failed requests.
    retry_backoff_ms: Optional[int] = None  # (Optional) Delay increases between retries in milliseconds.
    retry_base_ms: Optional[int] = None  # (Optional) Base retry delay in milliseconds.
    retry_ceiling_ms: Optional[int] = None  # (Optional) Maximum retry delay in milliseconds.
    req_timeout_millis: Optional[int] = None  # (Optional) Request timeout in milliseconds.
    stop_on_head: bool = False  # (Optional) Whether to automatically stop when reaching the blockchain head or keep the pipeline running indefinitely.
    head_poll_interval_millis: Optional[int] = None  # (Optional) How frequently (in milliseconds) to poll the blockchain head for updates.
    buffer_size: Optional[int] = None  # (Optional) Determines how many responses store in a buffer before sending them to the consumer.
    # RPC-specific fields
    compute_units_per_second: Optional[int] = None  # (Optional) Rate limit in compute units per second.
    batch_size: Optional[int] = None  # (Optional) Number of blocks fetched per batch.
    reorg_safe_distance: Optional[int] = None  # (Optional) Number of blocks behind head considered reorg-safe.
    trace_method: Optional[str] = None  # (Optional) Trace method: "trace_block" or "debug_trace_block_by_number".


class ResponseStream:
    def __init__(self, inner):
        self.inner = inner

    def close(self):
        self.inner.close()

    async def next(self) -> Optional[Dict[str, pyarrow.RecordBatch]]:
        return await self.inner.next()


def start_stream(cfg: ProviderConfig, query: Query) -> ResponseStream:
    # PyO3 extracts enum fields via Python's str(), which in Python 3.13+ returns
    # "ProviderKind.RPC" instead of "rpc" for StrEnum members. Normalize to .value.
    if isinstance(cfg.kind, Enum):
        cfg = dataclasses.replace(cfg, kind=cfg.kind.value)
    if isinstance(query.kind, Enum):
        query = dataclasses.replace(query, kind=query.kind.value)
    inner = cc.ingest.start_stream(cfg, query)
    return ResponseStream(inner)
